import json
from tabulate import tabulate
import numpy as np

if __name__ == "__main__":

    import argparse as ap
    listlen = None

    def comma_separated_floatlist(arg: str):
        values = arg.split(",")

        global listlen

        if listlen is None:
            listlen = len(values)
        else:
            if listlen != len(values):
                raise ap.ArgumentTypeError("Received lists of different lengths")

        return np.array([float(v) for v in values])

    def comma_separated_stringlist(arg: str or None):

        global listlen

        if arg is None and listlen is None:
            return np.array(["Unnamed" for _ in range(100)])
        elif arg is None and listlen is not None:
            return np.array(["Unnamed" for _ in range(listlen)])

        values = arg.split(",")
        return np.array([str(v) for v in values])


    parser = ap.ArgumentParser(description='ISA Calculator', formatter_class=ap.ArgumentDefaultsHelpFormatter)

    settings_group = parser.add_argument_group("Program Settings", description="Change the program actions")
    settings_group.add_argument("-a", "--altitude", help="Altitude of single calculation", type=float)
    settings_group.add_argument("-t", "--table", help="Generate a table (start, stop, step)", nargs=3, type=float, action='store')

    table_group = parser.add_argument_group("Table Settings", description="Change the tabulating properties")
    table_group.add_argument("--params", help="Comma-separated list of parameters to tabulate. Options: T/P/D/A/M or 'all'", default='all')
    table_group.add_argument("-o", "--outfile", help="Output result to file", type=str)
    table_group.add_argument("--print", nargs='?', const='psql', help="Print table to console given a format as specified by the tabulate package", type=str, default=None)

    atmos_group = parser.add_argument_group("Atmosphere Settings", description="Change the atmospheric properties")
    atmos_group.add_argument("--infile", help="Path to JSON file containing atmospheric description", default=None)
    atmos_group.add_argument("--temps", help="List of temperatures (requires --heights)", default=None, type=comma_separated_floatlist)
    atmos_group.add_argument("--altitudes", help="List of altitudes (requires --temps)", default=None, type=comma_separated_floatlist)
    atmos_group.add_argument("--names", help="List of layer names (requires --heights and --temps)", default=None, type=comma_separated_stringlist)
    atmos_group.add_argument("-p", "--p0", help="Base Pressure in Pa", default=None, type=float)
    atmos_group.add_argument("-d", "--d0", help="Base Density in kg/m^3", default=None, type=float)
    atmos_group.add_argument('-g', '--g0', help="Gravitational acceleration in m/s^2", type=float, default=9.80665)
    atmos_group.add_argument('-R', '--R', help="Gas constant for air in J/(kg*K)", type=float, default=287.0)
    atmos_group.add_argument('-y', '--gamma', help="Specific heat capacity ratio for air", type=float, default=1.4)

    args = parser.parse_args()

    from isacalc import Atmosphere

    # loading from JSON file
    if args.infile is not None:
        with open(args.file, "r") as file:
            atmos_dict = json.load(file)
        atmosphere = Atmosphere(**atmos_dict)

    # loading from cmd args
    elif (args.p0 is not None
        or args.d0 is not None
        or args.temps is not None
        or args.altitudes is not None
        or args.g0 is not None
        or args.R is not None
        or args.gamma is not None):

        atmosphere = Atmosphere(
            p0=args.p0, d0=args.d0,
            Hn=args.altitudes, Tn=args.temps, Nn=args.names,
            g0=args.g0, R=args.R, gamma=args.gamma
        )

    # default
    else:
        atmosphere = Atmosphere()

    if args.table:
        start, stop, step = args.table
        if None in args.table:
            raise ap.ArgumentTypeError("start, stop and step need to all be defined to tabulate")

        table = atmosphere.tabulate(start, stop, step, args.outfile, args.params)

        if args.print is not None:
            print(tabulate(table, headers = 'keys', tablefmt = args.print))

    elif args.altitude:
        res = atmosphere.calculate(args.altitude)
        for val in res:
            print(val, end=", ")
        print()


from copy import copy
from typing import List, Set, Union, Iterable

import numpy as np
import sys, os, json

import pandas as pd

from .layers import NormalLayer, IsothermalLayer, LayerType

param_names = {
    'T': 'Temperature [K]',
    'P': 'Pressure [Pa]',
    'D': 'Density [kg/m^3]',
    'A' : 'Speed of Sound [m/s]',
    'M' : 'Dynamic Viscosity [kg/(m*s)]'
}

default_atmosphere = None

import importlib.resources
with importlib.resources.open_text("isacalc", "isa.json") as file:
    default_atmosphere = json.load(file)




class Atmosphere(object):

    def __init__(self, *args, **kwargs):
        """
        Atmosphere Object.

        :keyword g0     :   gravitational acceleration in m/s^2
        :keyword R      :   universal gas constant for air in J/(kg*K)
        :keyword gamma  :   specific-heat ratio for air
        :keyword p0     :   base pressure in Pa
        :keyword d0     :   base density in kg / m^3
        :keyword Tn     :   list of temperatures in K
        :keyword Hn     :   heights corresponding to given temperatures
        :keyword Nn     :   list of layer names

        :keyword infile :   path to JSON file containing atmospheric data
        """

        self.__g0 = None
        self.__R = None
        self.__gamma = None

        self.__p0 = None
        self.__d0 = None
        self.__Nn = None
        self.__Tn = None
        self.__Hn = None

        self.__is_standard_atmosphere = True

        classname = type(self).__name__
        infile = kwargs.get('infile', None)
        if infile is not None:
            self.__is_standard_atmosphere = False
            with open(infile, "r") as file:
                atmosphere = json.load(file)
            for key in atmosphere.keys():
                setattr(self, f"_{classname}__{key}", atmosphere[key])

        else:
            for key in default_atmosphere.keys():
                if key in kwargs and kwargs[key] is not None:
                    self.__is_standard_atmosphere = False
                    setattr(self, f"_{classname}__{key}", kwargs[key])
                elif key in kwargs and kwargs[key] is None:
                    setattr(self, f"_{classname}__{key}", default_atmosphere[key])
                else:
                    if key in kwargs and key == 'Nn' and kwargs[key] is None:
                        setattr(self, f"_{classname}__{key}", "Noname")
                    else:
                        setattr(self, f"_{classname}__{key}", default_atmosphere[key])

        self.__Lt = self.__get_lapse(self.__Hn, self.__Tn)

        self.__layers = []
        self.__build()
        self.__last_idx = 0


    def __len__(self):
        """
        :return: Number of layers
        """
        return len(self.__layers)

    def __getitem__(self, idx):
        """
        Return Layer object at idx
        :param idx: index
        :return: layer object
        """
        return self.__layers[idx]

    def __str__(self):
        """
        Used for pretty-printing object info
        :return:
        """
        res = f"{'Standard' if self.__is_standard_atmosphere else 'Custom'} Atmosphere ({len(self)} layers):"
        res += f"\n * g0 = {self.__g0:<10.6f} [m/s^2]"
        res += f"\n * R  = {self.__R:<10.6f} [J/(kg*K)]"
        res += f"\n * y  = {self.__gamma:<10.6f} [-]"
        res += f"\n * p0 = {self.__p0:<10.3f} [Pa]"
        res += f"\n * d0 = {self.__d0:<10.6f} [kg/m^3]"
        for layer in self.__layers:
            res += '\n' + str(layer)
        return res

    def __repr__(self):
        """
        Used for pretty-printing object info
        :return:
        """
        return self.__str__()

    def calculate(self, h: float, *args, **kwargs) -> np.ndarray:
        """
        Calculate all atmospheric parameters at the given height
        :param h:               Altitude in meters
        :keyword start_index:   Start searching for the correct layer from given index.
                                Can be used to speed up calculations by skipping search.
                                Can result in ValueError if the relevant layer is skipped.
        :return:    Numpy Array containing Temp, Press, Dens, SoundSpeed, DynVisc
        """

        start_idx = kwargs.get('start_index', 0)

        if h > self.__Hn[-1] or h < self.__Hn[0]:
            raise ValueError("Height is out of bounds")

        for self.__last_idx in range(start_idx, len(self)):

            if abs(h - self.__Hn[self.__last_idx +1]) < 1e-3:
                return self.__layers[self.__last_idx ].ceiling_values

            elif abs(h - self.__Hn[self.__last_idx ]) < 1e-3:
                return self.__layers[self.__last_idx ].base_values

            elif self.__Hn[self.__last_idx ] < h < self.__Hn[self.__last_idx  + 1]:
                return self.__layers[self.__last_idx ].get_values_at(h)

            elif h > self.__Hn[self.__last_idx  + 1]:
                continue

        raise ValueError("Failed to calculate")

    def tabulate(self, start: float, stop: float, step: float, export_as: str = None, params: Iterable[str] = None, fastcalc: bool = True) -> pd.DataFrame:
        """
        Generate a table of atmospheric data
        :param start:       Table height start in meters
        :param stop:        Table height stop in meters (inclusive)
        :param step:        Table resolution in meters
        :param export_as:   Write table to csv/xlsx file
        :param params:      Set of parameters to tabulate with height (T/P/D/A/M or all)
        :param fastcalc:    Speed up the iterations by caching the last layer index
        :return:            Dataframe with the atmospheric table
        """

        if start is None or stop is None or step is None:
            raise RuntimeError("start, stop and step cannot be None")

        if params is None or (isinstance(params, str) and params.lower() == 'all'):
            params = ['T', 'P', 'D', 'A', 'M']
        else:
            if isinstance(params, list) or (isinstance(params, str) and params.lower() != 'all'):
                params = [p.upper() for p in params]

        heights = np.arange(start=start, stop=stop+step, step=step)
        table_shape = (len(heights), len(params)+1)
        result = np.zeros(table_shape, dtype=float)

        if fastcalc:
            self.__last_idx = 0

        for idx, height in enumerate(heights):

            if fastcalc:
                parameters = self.calculate(height, start_index=self.__last_idx)
            else:
                parameters = self.calculate(height)

            result[idx,0] = height
            result[idx,1:] = parameters

        df_result = pd.DataFrame(data=result,
                                 index=range(0, table_shape[0]),
                                 columns=['Height [m]'] + [param_names[p] for p in params])

        if export_as is not None:
            extension = export_as.split('.')[-1].lower()

            if extension == 'csv':
                df_result.to_csv(export_as, index_label='Index')

            else:
                df_result.to_excel(export_as, index_label='Index')

        return df_result

    @property
    def temperatures(self) -> np.ndarray:
        """
        Array of temperatures along layers
        :return:
        """
        return np.array(self.__Tn)

    @property
    def altitudes(self) -> np.ndarray:
        """
        Array of altitudes of layers
        :return:
        """
        return np.array(self.__Hn)

    @property
    def base_values(self) -> dict:
        """
        Dictionary with the atmospheric base values
        :return:
        """
        return {
            'p0': self.__p0,
            'd0': self.__d0,
            'T0': self.__Tn[0]
        }

    @property
    def constants(self) -> dict:
        """
        Dictionary of physical constants of atmosphere
        :return:
        """
        return {
            'g0': self.__g0,
            'R': self.__R,
            'gamma': self.__gamma
        }

    @property
    def height_boundaries(self):
        """
        Method to calculate for which range the atmosphere model can be used
        :return: Min, Max Height
        """
        return self.__Hn[0], self.__Hn[-1]

    @staticmethod
    def __get_lapse(Hn, Tn) -> list:
        """
        Static Method to calculate the layer types of all layers
        :param Hn: Heights
        :param Tn: Temperatures
        :return: Layer Types
        """

        types = []

        for i in range(len(Hn)-1):
            delta_T = Tn[i+1] - Tn[i]

            lapse = delta_T/(Hn[i+1] - Hn[i])

            if lapse != 0:
                if abs(delta_T) > 0.5:
                    types.append(LayerType.STANDARD)

                else:
                    types.append(LayerType.ISOTHERMAL)

            elif lapse == 0:
                types.append(LayerType.ISOTHERMAL)

        return types

    def __build(self) -> None:
        """
        Helper method to build the atmosphere object
        """

        p0, d0 = self.__p0, self.__d0

        for name, h0, h_i, T0, T_i, layer_type in zip(self.__Nn, self.__Hn[:-1], self.__Hn[1:], self.__Tn[:-1], self.__Tn[1:], self.__Lt):

            if layer_type == LayerType.STANDARD:

                layer = NormalLayer(base_height=h0,
                                    base_temperature=T0,
                                    base_pressure=p0,
                                    base_density=d0,
                                    max_height=h_i,
                                    top_temperature=T_i,
                                    name=name,
                                    g0=self.__g0,
                                    R=self.__R,
                                    gamma=self.__gamma)

                T0, p0, d0, a0, mu0 = layer.ceiling_values

            elif layer_type == LayerType.ISOTHERMAL:

                layer = IsothermalLayer(base_height=h0,
                                        base_temperature=T0,
                                        base_pressure=p0,
                                        base_density=d0,
                                        max_height=h_i,
                                        name=name,
                                        g0=self.__g0,
                                        R=self.__R,
                                        gamma=self.__gamma)

                T0, p0, d0, a0, mu0 = layer.ceiling_values

            else:
                raise ValueError

            self.__layers.append(layer)

        self.__layers = np.array(self.__layers)



import sys
from copy import copy
from typing import List, Tuple, Dict, Union, Set, Optional
from enum import Enum

import numpy as np


class LayerType(Enum):
    STANDARD    = 1
    ISOTHERMAL  = 2


class Layer(object):
    """
    Template Object for Layers
    """

    def __init__(self, base_height: float, base_temperature: float, base_pressure: float, base_density: float,
                 max_height: float, name: str = '', g0: float = 9.80665, R: float = 287.0, gamma: float = 1.4, *args, **kwargs):
        """
        All Properties of a layer, excluding its type, which will be expressed
        by making separate objects for the different layers
        :param base_height:         Height at which the layer starts
        :param base_temperature:    Temperature at the bottom of the layer
        :param base_pressure:       Pressure at the bottom of the layer
        :param base_density:        Density at the bottom ofthe layer
        :param max_height:          Height up to which the layer extends
        """

        self.g0 = g0
        self.R = R
        self.gamma = gamma

        self.__h0 = base_height
        self.__T0 = base_temperature
        self.__p0 = base_pressure
        self.__d0 = base_density
        self.__h_max = max_height

        if name:
            self.__name = name
        else:
            self.__name = "unnamed"

    def __str__(self):
        """
        Used for pretty-printing object info
        :return:
        """
        return f"{self.__name:<16s} | {self.__h0:10.3f} - {self.__h_max:10.3f} [m] | {self.base_values[0]:6.2f} - {self.ceiling_values[0]:6.2f} [K]"

    @staticmethod
    def sutherland_viscosity(T, mu0=1.716e-5, T0=273.15, S=110.4) -> float:
        """
        Method to calculate the dynamic viscosity of air
        :param T:   Temperature at which to calculate
        :param mu0: Reference viscosity
        :param T0:  Reference temperature
        :param S:   Sutherland Temperature
        :return:    Viscosity of air in kg/(m*s)
        """
        return mu0 * np.power((T/T0), 1.5) * (T0 + S)/(T + S)


    def speed_of_sound(self, temp: float) -> float:
        """
        Method to calculate the speed of sound at a certain temperature
        :param temp: Temperature in K
        :return: speed of sound in m/s
        """
        return np.sqrt(self.gamma*self.R*temp)

    @property
    def base_values(self) -> np.ndarray:
        """
        Getter function to obtain the hidden layer states
        :return: List of all base values
        """
        return np.array([self.__T0, self.__p0, self.__d0, self.speed_of_sound(self.__T0), self.sutherland_viscosity(self.__T0)])

    @property
    def ceiling_height(self) -> float:
        """
        Getter function to obtain the maximum height of the layer
        :return: Maximum height
        """
        return self.__h_max

    @property
    def base_height(self) -> float:
        """
        Getter function to obtain the height at which the layer starts
        :return: Base Height
        """
        return self.__h0

    @property
    def name(self) -> str:
        """
        Method to return the name of the layer
        :return: name
        """
        return copy(self.__name)

    @property
    def ceiling_values(self) -> None:
        """
        This method will be overridden
        """
        raise NotImplementedError("This method should be overridden")

    def get_values_at(self, h) -> None:
        """
        This method will be overridden
        """
        raise NotImplementedError("This method should be overridden")


class IsothermalLayer(Layer):

    def __init__(self, base_height: float,
                 base_temperature: float,
                 base_pressure: float,
                 base_density: float,
                 max_height: float,
                 name: str = '',
                 **kwargs):

        super().__init__(base_height=base_height,
                         base_temperature=base_temperature,
                         base_pressure=base_pressure,
                         base_density=base_density,
                         max_height=max_height,
                         name=name,
                         **kwargs)

    @property
    def ceiling_values(self) -> np.ndarray:
        """
        Method to get the temperature, pressure and density at the ceiling of the layer
        :return: temperature, pressure, density, speed of sound
        """

        T0, P0, D0, a0, mu0 = self.base_values
        h = self.ceiling_height
        h0 = self.base_height

        P = P0 * np.exp(-self.g0 / (self.R * T0) * (h - h0))
        D = D0 * np.exp(-self.g0 / (self.R * T0) * (h - h0))

        a = self.speed_of_sound(T0)
        mu = self.sutherland_viscosity(T0)

        return np.array([T0, P, D, a, mu])

    def get_values_at(self, h) -> np.ndarray:
        """
        Method to get the temperature, pressure and density at height h, between the base and ceiling of the layer
        :param h: Height at which to evaluate the temperature, pressure, density
        :return: temperature, pressure, density, speed of sound
        """
        h_max = self.ceiling_height
        h0 = self.base_height

        ret = np.zeros((5,), dtype=float)

        if h > h_max:
            raise ValueError(f"Given height exceeds layer height: {round(h, 3)} > {round(h_max, 3)}")

        if h < h0:
            raise ValueError(f"Given height is too low for given layer:  {round(h, 3)} < {round(h0, 3)}")

        if h == h0:
            return np.array(self.base_values)

        T0, P0, D0, a0, mu0 = self.base_values

        P = P0 * np.exp(-self.g0 / (self.R * T0) * (h - h0))
        D = D0 * np.exp(-self.g0 / (self.R * T0) * (h - h0))

        a = self.speed_of_sound(T0)
        mu = self.sutherland_viscosity(T0)

        return np.array([T0, P, D, a, mu])


class NormalLayer(Layer):

    def __init__(self, base_height: float,
                 base_temperature: float,
                 base_pressure: float,
                 base_density: float,
                 max_height: float,
                 top_temperature: float,
                 name: str = '',
                 **kwargs):

        super().__init__(base_height=base_height,
                         base_temperature=base_temperature,
                         base_pressure=base_pressure,
                         base_density=base_density,
                         max_height=max_height,
                         name=name,
                         **kwargs)

        self.__T_top = top_temperature

    @property
    def ceiling_values(self) -> np.ndarray:
        """
        Method to get the temperature, pressure and density at the ceiling of the layer
        :return: temperature, pressure, density, speed of sound
        """

        T0, P0, D0, a0, mu0 = self.base_values
        h = self.ceiling_height
        h0 = self.base_height

        L = (self.__T_top - T0) / (h - h0)
        C = -self.g0 / (L * self.R)  # To Simplify and shorten code we define the following expression for the exponent

        P = P0 * (self.__T_top / T0) ** C
        D = D0 * (self.__T_top / T0) ** (C - 1)

        a = self.speed_of_sound(self.__T_top)
        mu = self.sutherland_viscosity(self.__T_top)

        return np.array([self.__T_top, P, D, a, mu])

    def get_values_at(self, h) -> np.ndarray:
        """
        Method to get the temperature, pressure and density at height h, between the base and ceiling of the layer
        :param h: Height at which to evaluate the temperature, pressure, density
        :return: temperature, pressure, density, speed of sound
        """

        h_max = self.ceiling_height
        h0 = self.base_height

        if h > h_max:
            raise ValueError

        if h == h0:
            return self.base_values

        T0, P0, D0, a0, mu0 = self.base_values

        L = (self.__T_top - T0) / (h_max - h0)
        C = -self.g0 / (L * self.R)  # To Simplify and shorten code we define the following expression for the exponent

        T = T0 + L*(h - h0)

        P = P0 * np.power((T / T0) , C)
        D = D0 * np.power((T / T0) , (C - 1))

        a = self.speed_of_sound(T)
        mu = self.sutherland_viscosity(T)

        return np.array([T, P, D, a, mu])




from distutils.core import setup


def readme():
  with open('README.md') as f:
    return f.read()


version = 'v1.3.3'

setup(
  name = 'isacalc',         # How you named your package folder (MyLib)
  packages = ['isacalc'],   # Chose the same as "name"
  package_data={'' : ['**/*.json']},
  version = version,      # Start with a small number and increase it with every change you make
  license='GNU GPLv3',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Standard International Atmosphere Calculator',   # Give a short description about your library
  long_description=readme(),
  long_description_content_type='text/markdown',
  author = 'Luke de Waal',                   # Type in your name
  author_email = 'lr.de.waal.01@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/LukeDeWaal/ISA_Calculator',   # Provide either the link to your github or to your website
  download_url = f'https://github.com/LukeDeWaal/ISA_Calculator/archive/{version}.tar.gz',    # I explain this later on
  keywords = ['ISA','Aerospace','Aeronautical','Atmosphere'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy', 'pandas', 'tabulate'
      ],
  include_package_data=True,
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',   # Again, pick a license
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
  ],
)
